#! /usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Plugin to generate Sharp AC IR commands as done by the CRMC-B028JBEZ
#
# This module  is in part based on the work/code from:
#      ToniA      https://github.com/adafruit/Raw-IR-decoder-for-Arduino/pull/3/commits/887ed4204711c0b911571f3090b7fd066e93f006
#
# Copyright (c) 2023 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
##

import struct

from .hvaclib import HVAC, GenPluginObject


class Sharp(HVAC):
    """Generic Sharp HVAC object. It must have, at the very minimum
    "mode" and "temperature" capabilities"""

    STARTFRAME = [3800, 1900]
    ENDFRAME = [435, 10000]
    MARK = 435
    SPACE0 = 435
    SPACE1 = 1400

    def __init__(self):
        super().__init__()
        self.brand = "Sharp"
        self.model = "Generic"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "dry"],
            "temperature": [x for x in range(18, 38)],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {}
        self.status = {"mode": "cool", "temperature": 25}

        self.to_set = {}
        # Specify wether the bits order has to be swapped
        self.is_msb = True
        self.FBODY = b"\xaa\x5a\xcf\x10\x00\x00\x00\x00\x00\x80\x00\xe0"
        self.crc_special = 0x01

    def set_temperature(self, temp):
        if temp < self.capabilities["temperature"][0]:
            temp = self.capabilities["temperature"][0]
        elif temp > self.capabilities["temperature"][-1]:
            temp = self.capabilities["temperature"][-1]
        if self.status["temperature"] != temp:
            self.to_set["temperature"] = temp

    def code_temperature(self):
        if "temperature" in self.to_set:
            temp = self.to_set["temperature"]
            if "mode" in self.to_set and self.to_set["mode"] != "cool":
                temp = 17
            elif self.status["mode"] != "cool":
                temp = 17
            else:
                temp = self.status["temperature"]
        else:
            temp = self.status["temperature"]
        mask = bytearray(b"\x00" * len(self.FBODY))
        mask[6] = temp - 17
        return mask, False

    def set_fan(self, mode):
        if "fan" not in self.capabilities:
            return
        if mode not in self.capabilities["fan"]:
            return
        if self.status["fan"] != mode:
            self.to_set["fan"] = mode

    def code_fan(self):
        """mode is one of auto, lowest, low, middle, high, highest"""
        if "mode" in self.to_set and self.to_set["mode"] == "dry":
            mode = "auto"
            self.to_set["fan"] = mode
        elif self.status["mode"] == "dry":
            mode = "auto"
            self.to_set["fan"] = mode
        else:
            if "fan" in self.to_set:
                mode = self.to_set["fan"]
            else:
                if "fan" in self.capabilities:
                    mode = self.status["fan"]
                else:
                    mode = None
        rank = ["auto", "low", "lowest", "middle", "high", "highest"]
        mask = bytearray(b"\x00" * len(self.FBODY))
        if mode:
            fval = (rank.index(mode) + 2) << 4
            mask[6] = fval
        return mask, False

    def set_swing(self, mode):
        if "swing" not in self.capabilities:
            return
        if mode not in self.capabilities["swing"]:
            return
        if self.status["swing"] != mode:
            self.to_set["swing"] = mode

    def code_swing(self):
        if "swing" in self.to_set:
            mode = self.to_set["swing"]
        else:
            if "swing" in self.capabilities:
                mode = self.status["swing"]
            else:
                mode = False

        rank = ["auto", "ceiling", "90", "60", "45", "30" "swing"]
        if mode not in rank:
            mode = "auto"  # Just in case
        mask = bytearray(b"\x00" * len(self.FBODY))
        fval = rank.index(mode) + 8
        mask[8] = fval
        return mask, False

    def set_powerfull(self, mode="off"):
        # print("\n\nSharp set powerfull {}\n\n".format(mode))
        if "powerfull" not in self.status:
            return

        if "powerfull" in self.capabilities:
            checkwith = self.capabilities
        elif "powerfull" in self.xtra_capabilities:
            checkwith = self.xtra_capabilities
        # print("Setting powerfull as {} from {}".format(mode,checkwith))
        if mode not in checkwith["powerfull"]:
            return
        if self.status["powerfull"] != mode:
            self.to_set["powerfull"] = mode

    def code_powerfull(self, toggle=False):
        b10code = 0x00
        if toggle:
            if "powerfull" in self.to_set:
                b10code = 0x01

        mask = bytearray(b"\x00" * len(self.FBODY))
        mask[10] = b10code
        # Nothing to set... I know... it's the mode that will do this
        return mask, False

    def set_economy(self, mode="off"):
        if "economy" not in self.status:
            return

        if "exonomy" in self.capabilities:
            checkwith = self.capabilities
        elif "economy" in self.xtra_capabilities:
            checkwith = self.xtra_capabilities

        if mode not in checkwith["economy"]:
            return
        if self.status["economy"] != mode:
            self.to_set["economy"] = mode

    def code_economy(self, toggle=False):
        if not toggle:
            mode = self.status["economy"]
        elif "economy" in self.to_set:
            mode = self.to_set["economy"]
        else:
            if "economy" not in self.status:
                mode = "off"
            else:
                mode = self.status["economy"]

        mask = bytearray(b"\x00" * len(self.FBODY))
        if mode == "on":
            mask[11] |= 0x10
        return mask, False

    def set_purifier(self, mode="off"):
        if "purifier" not in self.capabilities:
            return
        if mode not in self.capabilities["purifier"]:
            return
        if self.status["purifier"] != mode:
            self.to_set["purifier"] = mode

    def code_purifier(self):

        mask = bytearray(b"\x00" * len(self.FBODY))

        return mask, False

    def set_mode(self, mode):
        if mode not in self.capabilities["mode"]:
            mode = "cool"
        if mode != self.status["mode"]:
            self.to_set["mode"] = mode

    def code_mode(self, option=None):
        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        mask = bytearray(b"\x00" * len(self.FBODY))
        if option is None or mode == "off":
            if mode != self.status["mode"]:
                if mode == "off":
                    b5val = 0x21
                    b6mode = self.status["mode"]
                elif self.status["mode"] == "off":
                    b5val = 0x11
                    b6mode = mode
                else:
                    b5val = 0x31
                    b6mode = mode
            else:
                b5val = 0x31
                b6mode = mode if mode != "off" else "cool"
        else:
            # Caller should not send option if off
            if option == "on":
                b5val = 0x61
            else:
                b5val = 0x71
            b6mode = mode
        mask[5] = b5val
        mask[6] = ["auto", "heat", "cool", "dry"].index(b6mode)
        return mask, False

    def build_code(self, withmode=True):
        packet = bytearray(self.FBODY)
        # Note that set mod must be last for it replaces values
        if withmode:
            lof = [
                self.code_temperature,
                self.code_fan,
                self.code_swing,
                self.code_purifier,
                self.code_economy,
                self.code_mode,
            ]
        else:
            lof = [
                self.code_temperature,
                self.code_fan,
                self.code_swing,
                self.code_purifier,
                self.code_economy,
            ]
        for f in lof:
            mask, replace = f()
            if replace:
                packet = bytearray([y or x for x, y in zip(packet, mask)])
            else:
                packet = bytearray([x | y for x, y in zip(packet, mask)])
        return packet

    def _build_ircode(self):
        frames = []
        normalframe = False
        for x in self.capabilities:
            if x in self.to_set:
                normalframe = True
                break
        if normalframe:
            frames += [self.build_code()]

        if ("mode" in self.to_set and (self.to_set["mode"] != "off")) or self.status[
            "mode"
        ] != "off":
            for prop in self.xtra_capabilities:
                # print("Looking at {} with {} and {}".format(prop,self.to_set,self.status))
                if prop in self.to_set and self.to_set[prop] != self.status[prop]:
                    f = getattr(self, "code_" + prop, None)
                    if f:
                        packet = self.build_code(withmode=False)
                        for mask, replace in [
                            self.code_mode(option=self.to_set[prop]),
                            f(toggle=True),
                        ]:
                            if replace:
                                packet = bytearray(
                                    [y or x for x, y in zip(packet, mask)]
                                )
                            else:
                                packet = bytearray(
                                    [x | y for x, y in zip(packet, mask)]
                                )
                        frames.append(packet)
        else:
            # We are off, so xtra_capabilities should also be off
            for x in self.xtra_capabilities:
                self.to_set[x] = "off"
        idx = 0
        for x in frames:
            frames[idx] += self.crc(x)
            idx += 1
        return frames

    def crc(self, frame):
        crc = 0
        for x in frame:
            crc ^= x
        crc ^= self.crc_special
        crc ^= crc >> 4
        crc = (crc & 0x0F) << 4
        crc += self.crc_special
        return crc.to_bytes(1, "big")

    def get_timing(self):
        # Well Sharp is different
        return {
            "start frame": self.STARTFRAME,
            "end frame": self.ENDFRAME,
            "mark": self.MARK,
            "space 0": self.SPACE0,
            "space 1": self.SPACE1,
        }


class JTech(Sharp):
    def __init__(self):
        super().__init__()
        self.model = "FTM-PV2S"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "dry"],
            "temperature": [x for x in range(14, 30)],
            "fan": ["auto", "highest", "middle", "low", "lowest"],
            "swing": ["auto", "ceiling", "90", "60", "45", "30", "swing"],
            "hswing": ["left", "middle", "right", "swing"],
            "target": [
                "off",
                "close left",
                "close middle",
                "close right",
                "far left",
                "far middle",
                "far right",
            ],
            "purifier": ["off", "on"],
        }
        self.xtra_capabilities = {"powerfull": ["off", "on"], "economy": ["off", "on"]}
        self.status = {
            "mode": "off",
            "temperature": 25,
            "fan": "auto",
            "swing": "auto",
            "hswing": "middle",
            "target": "off",
            "purifier": "off",
            "economy": "off",
            "powerfull": "off",
        }
        self.temperature_step = 0.5

    def code_purifier(self):

        mask = bytearray(b"\x00" * len(self.FBODY))

        if "purifier" in self.to_set:
            mode = self.to_set["purifier"]
        else:
            mode = self.status["purifier"]
        if mode == "on":
            mask[11] |= 0x04

        # Nothing to set... I know... it's the mode that will do this
        return mask, False

    def code_swing(self):
        """This one takes care of "swing", "horizontal swing" and "spot" (AKA target"""
        if ("target" in self.to_set and self.to_set["target"] == "off") or self.status[
            "target"
        ] == "off":
            # OK, swing and horizontal swing are on
            if "swing" in self.to_set:
                smode = self.to_set["swing"]
            else:
                smode = self.status["swing"]
            if "hswing" in self.to_set:
                hsmode = self.to_set["hswing"]
            else:
                hsmode = self.status["hswing"]
            b8val = (["", "middle", "left", "right"] + 11 * [""] + ["swing"]).index(
                hsmode
            ) << 4
            b8val += 8 + ["auto", "ceiling", "90", "60", "45", "30", "swing"].index(
                smode
            )
            b9val = 0x0
        else:
            # Target is on
            b9val = 0x01
            front, side = (
                ("target" in self.to_set and self.to_set["target"])
                or self.status["target"]
            ).split(" ")
            b8val = (["", "middle", "left", "right"]).index(side) << 4
            if front == "close":
                b8val += 0xC
            else:
                b8val += 0x9

        mask = bytearray(b"\x00" * len(self.FBODY))
        mask[9] |= b9val
        mask[8] = b8val
        return mask, False

    def set_hswing(self, mode):
        if "hswing" not in self.capabilities:
            return
        if mode not in self.capabilities["hswing"]:
            return
        if self.status["hswing"] != mode:
            self.to_set["hswing"] = mode

    def code_hswing(self):
        """Nothing to do. hNalded by code_swing"""
        return bytearray(b"\x00" * len(self.FBODY)), False

    def set_target(self, mode):
        if "target" not in self.capabilities:
            return
        if mode not in self.capabilities["target"]:
            return
        if self.status["target"] != mode:
            self.to_set["target"] = mode

    def code_target(self):
        """Nothing to do. hNalded by code_swing"""
        return bytearray(b"\x00" * len(self.FBODY)), False

    def code_temperature(self):
        if "temperature" in self.to_set:
            temp = self.to_set["temperature"]
        else:
            temp = self.status["temperature"]
        if "mode" in self.to_set:
            if self.to_set["mode"] != "cool":
                temp = None
        elif self.status["mode"] != "cool":
            temp = None
        mask = bytearray(b"\x00" * len(self.FBODY))
        if temp:
            if (temp * 10) % 10:
                deci = True
                temp = int(temp)
            else:
                deci = False
            if temp < 16:
                temp += 0x3E
                if deci:
                    temp += 0x20
            else:
                if not deci:
                    temp = 0xC0 + (temp - 15)
                else:
                    temp = 0x70 + (temp - 15)
            mask[4] = temp
        return mask, True


class PluginObject(GenPluginObject):
    MODELS = {"generic": Sharp, "j-tech": JTech}

    def __init__(self):
        self.brand = "sharp"


def main():

    import argparse
    import base64

    parser = argparse.ArgumentParser(description="Decode LIRC IR code into frames.")
    # version="%prog " + __version__ + "/" + bl.__version__)
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        default="generic",
        help="Set the A/C model. (default generic).",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="List known models and return.",
    )
    parser.add_argument(
        "-t", "--temp", type=int, default=25, help="Temperature (°C). (default 25)."
    )
    parser.add_argument(
        "-m",
        "--mode",
        choices=["auto", "cool", "dry", "off"],
        default="cool",
        help="Mode to set. (default 'cool').",
    )
    parser.add_argument(
        "-f",
        "--fan",
        choices=["auto", "highest", "high", "middle", "low", "lowest"],
        default="auto",
        help="Fan mode. (default 'auto').",
    )
    spotgroup = parser.add_mutually_exclusive_group()
    swinggroup = spotgroup.add_argument_group()
    swinggroup.add_argument(
        "-s",
        "--swing",
        choices=["auto", "ceiling", "90", "60", "45", "30", "swing"],
        default="auto",
        help="Set swing",
    )
    swinggroup.add_argument(
        "-H",
        "--hswing",
        choices=["middle", "left", "right", "swing"],
        default="middle",
        help="Set horizontal swing",
    )
    spotgroup.add_argument(
        "-S",
        "--spot",
        choices=[
            "off",
            "close left",
            "close middle",
            "close right",
            "far left",
            "far middle",
            "far right",
        ],
        default="off",
        help="Set spot",
    )
    parser.add_argument(
        "-p", "--powerfull", action="store_true", default=False, help="Set super jet"
    )
    parser.add_argument(
        "-P",
        "--plasma",
        action="store_true",
        default=False,
        help="Set Plasmacluster purifier",
    )
    parser.add_argument(
        "-e", "--economy", action="store_true", default=False, help="Set economy mode"
    )
    parser.add_argument(
        "-l",
        "--lirc",
        action="store_true",
        default=False,
        help="Output LIRC compatible timing",
    )
    parser.add_argument(
        "-b",
        "--broadlink",
        action="store_true",
        default=False,
        help="Output Broadlink timing",
    )
    parser.add_argument(
        "-B",
        "--base64",
        action="store_true",
        default=False,
        help="Output Broadlink timing in base64 encoded",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))

    if opts.list:
        print(f"Available modelas are: {PluginObject().MODELS.keys()}")

    device = PluginObject().factory(opts.model)
    frames = []
    device.set_temperature(opts.temp)
    device.set_fan(opts.fan)
    device.set_swing(opts.swing)
    device.set_hswing(opts.hswing)
    device.set_target(opts.spot)
    device.set_purifier((opts.plasma and "on") or "off")
    device.set_powerfull((opts.powerfull and "on") or "off")
    device.set_economy((opts.economy and "on") or "off")
    device.set_mode(opts.mode)

    frames = device.build_ircode()

    if opts.lirc:
        lircf = device.to_lirc(frames)
        while lircf:
            print("\t".join(["%d" % x for x in lircf[:6]]))
            lircf = lircf[6:]
    elif opts.broadlink or opts.base64:
        lircf = device.to_lirc(frames)
        bframe = device.to_broadlink([int(x) for x in lircf])
        if opts.base64:
            print("{}".format(str(base64.b64encode(bframe), "ascii")))
        else:
            print("{}".format(bframe.hex()))
    else:
        for f in frames:
            print(" ".join([hex(x) for x in f]))


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Plugin to generate LG AC IR commands as done by the AXB74515402
#
# Found the info about 4 bits somewhere on the  Internet...Can't find
# it again. Apologies for not being able to thanks that person.
#
# Copyright (c) 2023 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
##

import struct

from .hvaclib import HVAC, GenPluginObject


class LG(HVAC):
    """Generic LG HVAC object. It must have, at the very minimum
    "mode" and "temperature" capabilities"""

    STARTFRAME = [3100, 9850]
    ENDFRAME = [520, 12000]
    MARK = 520
    SPACE0 = 520
    SPACE1 = 1530

    def __init__(self):
        super().__init__()
        self.brand = "LG"
        self.model = "Generic"
        self.capabilities = {
            "mode": ["off", "cool", "fan", "dry"],
            "temperature": [x for x in range(18, 30)],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {}
        self.status = {"mode": "off", "temperature": 25}

        self.to_set = {}
        # Specify wether the bits order has to be swapped
        self.is_msb = False
        self.FBODY = b"\x88\x00\x00"

    def set_temperature(self, temp):
        if temp < self.capabilities["temperature"][0]:
            temp = self.capabilities["temperature"][0]
        elif temp > self.capabilities["temperature"][-1]:
            temp = self.capabilities["temperature"][-1]
        if temp != self.status["temperature"]:
            self.to_set["temperature"] = temp

    def code_temperature(self):
        if "temperature" in self.to_set:
            temp = self.to_set["temperature"]
        else:
            temp = self.status["temperature"]

        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        if mode == "fan":
            temp = 18
        elif mode == "auto":
            if "auto_bias" in self.status:
                try:  # Getting ahead of oneself...
                    if "auto_bias" in self.to_set:
                        temp = 15 + self.capabilities["auto_bias"].index(
                            self.to_set["auto_bias"]
                        )
                    else:
                        temp = 15 + self.capabilities["auto_bias"].index(
                            self.status["auto_bias"]
                        )
                except:
                    temp = 17
            else:
                temp = 17
        elif mode == "dry":
            temp = 24
        mask = bytearray(b"\x00" * 3)
        mask[2] = (temp - 15) << 4
        return mask, False

    def set_auto_bias(self, mode):
        if "auto_bias" not in self.capabilities:
            return
        if mode not in self.capabilities["auto_bias"]:
            return
        if self.status["auto_bias"] != mode:
            self.to_set["auto_bias"] = mode

    def set_fan(self, mode):
        if "fan" not in self.capabilities:
            return
        if mode not in self.capabilities["fan"]:
            return
        if self.status["fan"] != mode:
            self.to_set["fan"] = mode

    def code_fan(self):
        """mode is one of auto, lowest, low, middle, high, highest"""
        if "fan" in self.to_set:
            fmode = self.to_set["fan"]
        else:
            if "fan" in self.capabilities:
                fmode = self.status["fan"]
            else:
                fmode = "auto"
        rank = {
            "lowest": 0x0,
            "low": 0x09,
            "middle": 0x02,
            "high": 0x0A,
            "highest": 0x04,
            "auto": 0x05,
        }

        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        if mode == "auto":
            fmode = "auto"
        if fmode not in rank:
            fmode = "auto"  # Just in case
        mask = bytearray(b"\x00" * 3)
        mask[2] = rank[fmode]
        return mask, False

    def set_swing(self, mode):
        if "swing" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["swing"]:
            return
        if self.status["swing"] != mode:
            self.to_set["swing"] = mode

    def code_swing(self):
        if "swing" in self.to_set:
            mode = self.to_set["swing"]

            if mode in ["swing", "on"]:
                return bytearray(b"\x88\x13\x14")
            elif mode == "off":
                return bytearray(b"\x88\x13\x15")
            elif mode == "0":
                return bytearray(b"\x88\x13\x04")
            elif mode == "15":
                return bytearray(b"\x88\x13\x05")
            elif mode == "30":
                return bytearray(b"\x88\x13\x06")
            elif mode == "45":
                return bytearray(b"\x88\x13\x07")
            elif mode == "60":
                return bytearray(b"\x88\x13\x08")
            elif mode == "90":
                return bytearray(b"\x88\x13\x09")

        return bytearray()

    def set_hswing(self, mode):
        if "hswing" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["hswing"]:
            return
        if self.status["hswing"] != mode:
            self.to_set["hswing"] = mode

    def code_hswing(self):
        if "hswing" in self.to_set:
            mode = self.to_set["hswing"]

            if mode in ["swing", "on"]:
                return bytearray(b"\x88\x13\x16")
            elif mode == "off":
                return bytearray(b"\x88\x13\x17")
            elif mode == "left":
                return bytearray(b"\x88\x13\x0b")
            elif mode == "centre left":
                return bytearray(b"\x88\x13\x0c")
            elif mode == "centre":
                return bytearray(b"\x88\x13\x0d")
            elif mode == "centre right":
                return bytearray(b"\x88\x13\x0e")
            elif mode == "right":
                return bytearray(b"\x88\x13\x0f")
            elif mode == "swing left":
                return bytearray(b"\x88\x13\x10")
            elif mode == "swing right":
                return bytearray(b"\x88\x13\x11")
        return bytearray()

    def set_powerfull(self, mode="off"):
        # print("\n\nLG set powerfull {}\n\n".format(mode))
        if "powerfull" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["powerfull"]:
            return
        if self.status["powerfull"] != mode:
            self.to_set["powerfull"] = mode

    def code_powerfull(self):

        if "powerfull" in self.to_set:
            mode = self.to_set["powerfull"]
            if mode == "on":
                return bytearray(b"\x88\x10\x08")
            # Off... send the "normal" code
            dosend = True
            for x in self.capabilities:
                if x in self.to_set:
                    dosend = False
                    break
            if dosend:
                otoset = self.to_set
                self.to_set = {"mode": self.status["mode"]}
                frame = self.build_code()
                self.to_set = otoset
                return frame
        return []

    def set_purifier(self, mode="off"):
        if "purifier" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["purifier"]:
            return
        # This is a toggling value AFAIK
        if self.status["purifier"] != mode:
            self.to_set["purifier"] = mode

    def code_purifier(self):
        """Some LG AC seem to have such"""
        if self.to_set["purifier"] == "on":
            return bytearray(b"\x88\xc0\x00")
        else:
            return bytearray(b"\x88\xc0\x08")

    def set_economy(self, mode):
        if "economy" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["economy"]:
            return
        if self.status["economy"] != mode:
            self.to_set["economy"] = mode

    def code_economy(self):
        if "economy" in self.to_set:
            mode = self.to_set["economy"]

            if mode == "off":
                return bytearray(b"\x88\xc0\x7f")
            elif mode == "80":
                return bytearray(b"\x88\xc0\x7d")
            elif mode == "60":
                return bytearray(b"\x88\xc0\x7e")
            elif mode == "40":
                return bytearray(b"\x88\xc0\x80")

        return bytearray()

    def set_cleaning(self, mode):
        if "cleaning" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["cleaning"]:
            return
        if self.status["cleaning"] != mode:
            self.to_set["cleaning"] = mode

    def code_cleaning(self):
        if "cleaning" in self.to_set:
            mode = self.to_set["cleaning"]

            if mode == "off":
                return bytearray(b"\x88\xc0\x0b")
            elif mode == "on":
                return bytearray(b"\x88\xc0\x0c")

        return bytearray()

    def set_diagnostic(self):
        if "diagnostic" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["diagnostic"]:
            return
        if self.status["diagnostic"] != mode:
            self.to_set["diagnostic"] = mode

    def code_cleaning(self):
        if "diagnostic" in self.to_set:
            mode = self.to_set["diagnostic"]
            self.to_set["diagnostic"] = "off"  # It is a request, not a toggle
            if mode == "on":
                return bytearray(b"\x88\xc0\xce")

    def set_mode(self, mode):
        if mode not in self.capabilities["mode"]:
            mode = "cool"
        if mode != self.status["mode"]:
            self.to_set["mode"] = mode

    def code_mode(self):
        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        mask = bytearray(b"\x88\x00\x00")
        if mode == "off":
            return bytearray(b"\x00\xc0\x05"), True
        if self.status["mode"] != "off":
            addit = 8
        else:
            addit = 0
        mask = bytearray(b"\x00\x00\x00")
        if mode == "auto":
            mask[1] = 3 + addit
        elif mode == "cool":
            mask[1] = addit
        elif mode == "dry":
            mask[1] = 1 + addit
        elif mode == "fan":
            mask[1] = 2 + addit

        return mask, False

    def build_code(self):
        frames = []
        packet = bytearray(self.FBODY)
        # Note that set mod must be last for it replaces values
        if {"mode", "temperature", "fan"}.intersection(set(self.to_set.keys())):
            for f in [self.code_temperature, self.code_fan, self.code_mode]:
                mask, replace = f()
                if replace:
                    packet = bytearray([y or x for x, y in zip(packet, mask)])
                else:
                    packet = bytearray([x | y for x, y in zip(packet, mask)])
            frames += [packet]
        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]
        if mode != "off":
            print("Mode is {} from {}".format(mode, self.to_set))
            for prop in self.xtra_capabilities:
                # print("Looking at {} with {} and {}".format(prop,self.to_set,self.status))
                if prop in self.to_set and self.to_set[prop] != self.status[prop]:
                    f = getattr(self, "code_" + prop, None)
                    if f:
                        frames.append(f())
        return frames

    def _build_ircode(self):
        frames = []
        frames += self.build_code()
        idx = 0
        for x in frames:
            frames[idx] += self.crc(x)
            idx += 1
        return frames

    def crc(self, frame):
        crc = 0
        for x in frame:
            crc += (x & 0xF0) >> 4
            crc += x & 0x0F
        return ((crc & 0x0F) << 4).to_bytes(1, "big")

    def get_timing(self):
        # Well LG is different
        return {
            "start frame": STARTFRAME,
            "end frame": ENDFRAME,
            "mark": MARK,
            "space 0": SPACE0,
            "space 1": SPACE1,
            "drop_bits": 4,
        }


class InverterV(LG):
    def __init__(self):
        super().__init__()
        self.model = "Inverter V"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "fan", "dry"],
            "temperature": [x for x in range(16, 30)],
            "auto_bias": ["-2", "-1", "default", "+1", "+2"],
            "fan": ["auto", "highest", "high", "middle", "low", "lowest"],
        }
        self.xtra_capabilities = {
            "swing": ["off", "swing", "90", "0"],
            "powerfull": ["off", "on"],
            "cleaning": ["off", "on"],
            "economy": ["off", "80", "60", "40"],
        }
        self.status = {
            "mode": "off",
            "temperature": 25,
            "fan": "auto",
            "swing": "off",
            "auto_bias": "default",
            "powerfull": "off",
            "cleaning": "off",
            "economy": "off",
        }


class DualInverter(LG):
    def __init__(self):
        super().__init__()
        self.model = "Dual Imverter"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "fan", "dry"],
            "temperature": [x for x in range(16, 30)],
            "auto_bias": ["-2", "-1", "default", "+1", "+2"],
            "fan": ["auto", "highest", "high", "middle", "low", "lowest"],
        }
        self.xtra_capabilities = {
            "swing": ["off", "swing", "90", "60", "45", "30", "15", "0"],
            "hswing": [
                "off",
                "swing",
                "left",
                "centre left",
                "centre",
                "centre right",
                "right",
                "swing left",
                "swing right",
            ],
            "powerfull": ["off", "on"],
            "purifier": ["off", "on"],
            "cleaning": ["off", "on"],
            "economy": ["off", "80", "60", "40"],
            "diagnostic": ["off", "on"],
        }
        self.status = {
            "mode": "off",
            "temperature": 25,
            "fan": "auto",
            "swing": "off",
            "auto_bias": "default",
            "powerfull": "off",
            "purifier": "off",
            "cleaning": "off",
            "economy": "off",
            "diagnostic": "off",
        }


class PluginObject(GenPluginObject):

    MODELS = {"generic": LG, "inverter v": InverterV, "dual inverter": DualInverter}

    def __init__(self):
        self.brand = "lg"


def main():

    import argparse
    import base64

    parser = argparse.ArgumentParser(description="Generate LG A/C codes.")
    # version="%prog " + __version__ + "/" + bl.__version__)
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        default="generic",
        help="Set the A/C model. (default generic).",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="List known models and return.",
    )
    parser.add_argument(
        "-t", "--temp", type=int, default=25, help="Temperature (°C). (default 25)."
    )
    parser.add_argument(
        "-m",
        "--mode",
        choices=["auto", "cool", "dry", "fan", "off"],
        default="cool",
        help="Mode to set. (default 'cool').",
    )
    parser.add_argument(
        "-f",
        "--fan",
        choices=["auto", "highest", "high", "middle", "low", "lowest"],
        default="auto",
        help="Fan mode. (default 'auto').",
    )
    parser.add_argument(
        "-s",
        "--swing",
        choices=["off", "swing", "90", "0"],
        default="off",
        help="Swing mode. (default 'off').",
    )
    parser.add_argument(
        "-e",
        "--economy",
        choices=["off", "80", "60", "40"],
        default="off",
        help="Economy mode. (default 'off').",
    )
    parser.add_argument(
        "-A",
        "--bias",
        choices=["-2", "-1", "default", "1", "2"],
        default="default",
        help="Auto mode temperature bias. (default 'default').",
    )
    parser.add_argument(
        "-P", "--purifier", action="store_true", default=False, help="Set plasma"
    )
    parser.add_argument(
        "-p", "--powerfull", action="store_true", default=False, help="Set powerfull"
    )
    parser.add_argument(
        "-l",
        "--lirc",
        action="store_true",
        default=False,
        help="Output LIRC compatible timing",
    )
    parser.add_argument(
        "-b",
        "--broadlink",
        action="store_true",
        default=False,
        help="Output Broadlink timing",
    )
    parser.add_argument(
        "-B",
        "--base64",
        action="store_true",
        default=False,
        help="Output Broadlink timing in base64 encoded",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))

    if opts.list:
        print(f"Available modelas are: {PluginObject().MODELS.keys()}")

    device = PluginObject().factory(opts.model)
    frames = []
    device.set_temperature(opts.temp)
    device.set_fan(opts.fan)
    device.set_swing(opts.swing)
    device.set_powerfull((opts.powerfull and "on") or "off")
    device.set_auto_bias(opts.bias)
    device.set_economy(opts.economy)
    device.set_purifier((opts.purifier and "on") or "off")
    device.set_mode(opts.mode)

    frames = device.build_ircode()

    if opts.lirc:
        lircf = device.to_lirc(frames)
        while lircf:
            print("\t".join(["%d" % x for x in lircf[:6]]))
            lircf = lircf[6:]
    elif opts.broadlink or opts.base64:
        lircf = device.to_lirc(frames)
        bframe = device.to_broadlink([int(x) for x in lircf])
        if opts.base64:
            print("{}".format(str(base64.b64encode(bframe), "ascii")))
        else:
            print("{}".format(bframe.hex()))
    else:
        for f in frames:
            print(" ".join([hex(x) for x in f]))


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import argparse
import base64
import importlib
import sys
from pathlib import Path


def main():

    # Let's get the plugins
    dir_path = Path(__file__).parent / "plugins"
    plugs = {}
    for filename in dir_path.iterdir():
        if filename.name == "hvaclib.py":
            continue
        if filename.suffix == ".py" and not filename.stem.startswith("__"):
            mod = importlib.import_module(".plugins." + filename.stem, package="pyhvac")
            thisplug = mod.PluginObject()
            plugs[thisplug.brand] = thisplug

    parser = argparse.ArgumentParser(description="Decode LIRC IR code into frames.")
    # version="%prog " + __version__ + "/" + bl.__version__)
    parser.add_argument(
        "-c", "--manufacturer", type=str.lower, default=None, help="Specify the A/C brand."
    )
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        default="generic",
        help="Set the A/C model. (default generic).",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="List known models and return.",
    )
    parser.add_argument(
        "-t", "--temp", type=int, default=25, help="Temperature (°C). (default 25)."
    )
    parser.add_argument(
        "-m",
        "--mode",
        choices=["auto", "cool", "dry", "fan", "off"],
        default="cool",
        help="Mode to set. (default 'cool').",
    )
    parser.add_argument(
        "-f",
        "--fan",
        choices=["auto", "highest", "high", "middle", "low", "lowest"],
        default="auto",
        help="Fan mode. (default 'auto').",
    )
    parser.add_argument(
        "-s", "--swing", action="store_true", default=False, help="Set swing"
    )
    parser.add_argument(
        "-p", "--powerfull", action="store_true", default=False, help="Set powerfull"
    )
    parser.add_argument(
        "-e",
        "--economy",
        choices=["on", "off", "80", "60", "40"],
        default="off",
        help="Economy mode. (default 'off').",
    )
    parser.add_argument(
        "-n",
        "--filter",
        action="store_true",
        default=False,
        help="Particulate filter mode",
    )
    parser.add_argument(
        "-o", "--clean", action="store_true", default=False, help="Odour filter mode"
    )
    parser.add_argument(
        "-l",
        "--lirc",
        action="store_true",
        default=False,
        help="Output LIRC compatible timing",
    )
    parser.add_argument(
        "-b",
        "--broadlink",
        action="store_true",
        default=False,
        help="Output Broadlink timing",
    )
    parser.add_argument(
        "-B",
        "--base64",
        action="store_true",
        default=False,
        help="Output Broadlink timing in base64 encoded",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))

    if opts.list:
        lofm={}
        for m in plugs:
            lofm[m] = [x for x in plugs[m].MODELS.keys()]
        print(f"Available models are: {lofm}")
        sys.exit(0)

    if opts.manufacturer in plugs:
        device = plugs[opts.manufacturer].factory(opts.model)
    else:
        print(f"Error: Manufacturer {opts.manufacturer} is not supported.")
        sys.exit(2)

    frames = []
    device.set_value("set_temperature",opts.temp)
    device.set_value("set_fan",opts.fan)
    device.set_value("set_swing",opts.swing)
    device.set_value("set_powerfull",(opts.powerfull and "on") or "off")
    device.set_value("set_purifier",(opts.filter and "on") or "off")
    device.set_value("set_cleaning",(opts.clean and "on") or "off")
    device.set_value("set_economy",opts.economy)
    device.set_mode(opts.mode)

    frames = device.build_ircode()

    if opts.lirc:
        lircf = device.to_lirc(frames)
        while lircf:
            print("\t".join(["%d" % x for x in lircf[:6]]))
            lircf = lircf[6:]
    elif opts.broadlink or opts.base64:
        lircf = device.to_lirc(frames)
        bframe = device.to_broadlink([int(x) for x in lircf])
        if opts.base64:
            print("{}".format(str(base64.b64encode(bframe), "ascii")))
        else:
            print("{}".format(bframe.hex()))
    else:
        for f in frames:
            print(" ".join([hex(x) for x in f]))



#! /usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Plugin to generate Daikin AC IR commands as done by the ARC480A44
#
# This module  is in part based on the work/code from:
#      Scott Kyle https://gist.github.com/appden/42d5272bf128125b019c45bc2ed3311f
#      mat_fr     https://www.instructables.com/id/Reverse-engineering-of-an-Air-Conditioning-control/
#
# Copyright (c) 2023 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
##

import struct

from .hvaclib import HVAC, GenPluginObject, bit_reverse

class Daikin(HVAC):
    """Generic Daikin HVAC object. It must have, at the very minimum
    "mode" and "temperature" capabilities"""

    FBODY = b"\x88\x5b\xe4\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa3\x00\x10"

    def __init__(self):
        super().__init__()
        self.brand = "Daikin"
        self.model = "Generic"
        self.capabilities = {
            "mode": ["off", "cool", "fan", "dry"],
            "temperature": [x for x in range(18, 32)],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {}
        self.status = {"mode": "cool", "temperature": 25}

        self.to_set = {}
        # Specify wether the bits order has to be swapped
        self.is_msb = False

    def set_temperature(self, temp):
        if temp < self.capabilities["temperature"][0]:
            temp = self.capabilities["temperature"][0]
        elif temp > self.capabilities["temperature"][-1]:
            temp = self.capabilities["temperature"][-1]
        self.to_set["temperature"] = temp

    def code_temperature(self):
        if "temperature" in self.to_set:
            temp = self.to_set["temperature"]
            if "mode" in self.to_set and self.to_set["mode"] == "fan":
                temp = 25
        else:
            temp = self.status["temperature"]
        mask = bytearray(b"\x00" * 18)
        mask[6] = bit_reverse(temp * 2)
        return mask, False

    def set_fan(self, mode):
        if "fan" not in self.capabilities:
            return
        if mode not in self.capabilities["fan"]:
            return
        self.to_set["fan"] = mode

    def code_fan(self):
        """mode is one of auto, lowest, low, middle, high, highest"""
        if "fan" in self.to_set:
            mode = self.to_set["fan"]
        else:
            if "fan" in self.capabilities:
                mode = self.status["fan"]
            else:
                mode = "auto"
        rank = ["lowest", "low", "middle", "high", "highest"]
        if mode not in rank:
            mode = "auto"  # Just in case
        mask = bytearray(b"\x00" * 18)
        if mode == "auto":
            mask[8] = 0x05
        else:
            mask[8] = bit_reverse(48 + (16 * rank.index(mode)))
        return mask, False

    def set_swing(self, mode):
        if "swing" not in self.capabilities:
            return
        if mode not in self.capabilities["swing"]:
            return
        self.to_set["swing"] = mode

    def code_swing(self):
        if "swing" in self.to_set:
            mode = self.to_set["swing"]
        else:
            if "swing" in self.capabilities:
                mode = self.status["swing"]
            else:
                mode = False

        mask = bytearray(b"\x00" * 18)
        if mode == "on":
            mask[8] = 0xF0
        else:
            mask[8] = 0x00
        return mask, False

    def set_powerfull(self, mode="off"):
        # print("\n\nDaikin set powerfull {}\n\n".format(mode))
        if "powerfull" not in self.capabilities:
            return
        if mode not in self.capabilities["powerfull"]:
            return
        self.to_set["powerfull"] = mode

    def code_powerfull(self):

        if "powerfull" in self.to_set:
            mode = self.to_set["powerfull"]
        else:
            if "powerfull" in self.capabilities:
                mode = self.status["powerfull"]
            else:
                mode = False

        mask = bytearray(b"\x00" * 18)
        if mode == True:
            mask[13] = 0x80
        return mask, True

    def set_comfort(self, mode="off"):
        if "comfort" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["comfort"]:
            return
        # This is a toggling value AFAIK
        if self.status["comfort"] != mode:
            self.to_set["comfort"] = mode

    def code_comfort(self):
        """Some Daikin AC seem to have such"""
        return []

    def set_mode(self, mode):
        if mode not in self.capabilities["mode"]:
            mode = "cool"
        self.to_set["mode"] = mode
        if mode == "fan":
            self.to_set["temperature"] = 25
        if mode == "off":
            self.to_set = {"mode": "off"}

    def code_mode(self):
        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        mask = bytearray(b"\x00" * 18)
        if mode == "off":
            mask[16] = 0x02
            cmode = self.status["mode"]
            if cmode == "off":
                mask[5] = 0x0C
            elif cmode == "dry":
                mask[5] = 0x04
            elif cmode == "fan":
                mask[5] = 0x06
            elif cmode == "heat":
                mask[5] = 0x02
            elif cmode == "auto":
                mask[5] = 0x00
        elif mode == "dry":
            mask[5] = 0x84
            mask[6] = 0x03
        elif mode == "fan":
            mask[5] = 0x86
        elif mode == "heat":
            mask[5] = 0x82
        elif mode == "auto":
            mask[5] = 0x80
        else:
            mask[5] = 0x8C
        return mask, True

    def build_code(self):
        frames = []
        packet = bytearray(self.FBODY)
        # Note that set mod must be last for it replaces values
        for f in [
            self.code_temperature,
            self.code_fan,
            self.code_swing,
            self.code_powerfull,
            self.code_mode,
        ]:
            mask, replace = f()
            if replace:
                packet = bytearray([y or x for x, y in zip(packet, mask)])
            else:
                packet = bytearray([x | y for x, y in zip(packet, mask)])
        frames += [packet]
        return frames

    def _build_ircode(self):
        frames = []
        frames += self.code_comfort()
        frames += self.build_code()
        idx = 0
        for x in frames:
            frames[idx] += self.crc(x)
            idx += 1
        return frames

    def crc(self, frame):
        crc = 0
        for x in frame:
            # print("Adding 0x%02x as 0x%02x"%(x,bit_reverse(x)))
            crc += bit_reverse(x)
            # print("crc now 0x%02x"%crc)
        return bit_reverse(crc & 0xFF).to_bytes(1, "big")


class Smash2(Daikin):
    def __init__(self):
        super().__init__()
        self.model = "Smash 2"
        self.capabilities = {
            "mode": ["off", "cool", "fan", "dry"],
            "temperature": [x for x in range(18, 32)],
            "fan": ["auto", "highest", "high", "middle", "low", "lowest"],
            "swing": ["off", "on"],
            "powerfull": ["off", "on"],
        }
        self.status = {
            "mode": "off",
            "temperature": 25,
            "fan": "auto",
            "swing": "off",
            "powerfull": "off",
        }


class PluginObject(GenPluginObject):
    MODELS = {"generic": Daikin, "smash 2": Smash2}

    def __init__(self):
        self.brand = "daikin"


def main():

    import argparse
    import base64

    parser = argparse.ArgumentParser(description="Decode LIRC IR code into frames.")
    # version="%prog " + __version__ + "/" + bl.__version__)
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        default="generic",
        help="Set the A/C model. (default generic).",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="List known models and return.",
    )
    parser.add_argument(
        "-t", "--temp", type=int, default=25, help="Temperature (°C). (default 25)."
    )
    parser.add_argument(
        "-m",
        "--mode",
        choices=["cool", "dry", "fan", "off"],
        default="cool",
        help="Mode to set. (default 'cool').",
    )
    parser.add_argument(
        "-f",
        "--fan",
        choices=["auto", "highest", "high", "middle", "low", "lowest"],
        default="auto",
        help="Fan mode. (default 'auto').",
    )
    parser.add_argument(
        "-s", "--swing", action="store_true", default=False, help="Set swing"
    )
    parser.add_argument(
        "-p", "--powerfull", action="store_true", default=False, help="Set powerfull"
    )
    parser.add_argument(
        "-l",
        "--lirc",
        action="store_true",
        default=False,
        help="Output LIRC compatible timing",
    )
    parser.add_argument(
        "-b",
        "--broadlink",
        action="store_true",
        default=False,
        help="Output Broadlink timing",
    )
    parser.add_argument(
        "-B",
        "--base64",
        action="store_true",
        default=False,
        help="Output Broadlink timing in base64 encoded",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))

    if opts.list:
        print(f"Available modelas are: {PluginObject().MODELS.keys()}")

    device = PluginObject().factory(opts.model)

    frames = []
    device.set_temperature(opts.temp)
    device.set_fan(opts.fan)
    device.set_swing(opts.swing)
    device.set_powerfull(opts.powerfull)
    device.set_mode(opts.mode)

    frames = device.build_ircode()

    if opts.lirc:
        lircf = device.to_lirc(frames)
        while lircf:
            print("\t".join(["%d" % x for x in lircf[:6]]))
            lircf = lircf[6:]
    elif opts.broadlink or opts.base64:
        lircf = device.to_lirc(frames)
        bframe = device.to_broadlink([int(x) for x in lircf])
        if opts.base64:
            print("{}".format(str(base64.b64encode(bframe), "ascii")))
        else:
            print("{}".format(bframe.hex()))
    else:
        for f in frames:
            print(" ".join([hex(x) for x in f]))


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Plugin to generate Panasonic AC IR commands
#
# This module without the work/code from:
#      Scott Kyle https://gist.github.com/appden/42d5272bf128125b019c45bc2ed3311f
#      mat_fr     https://www.instructables.com/id/Reverse-engineering-of-an-Air-Conditioning-control/
#      user two, mathieu, vincent
#                 https://www.analysir.com/blog/2014/12/27/reverse-engineering-panasonic-ac-infrared-protocol/
#
# Copyright (c) 2023 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
##

import struct

from hashlib import blake2b


def bit_reverse(i, n=8):
    return int(format(i, "0%db" % n)[::-1], 2)


class HVAC(object):
    # 90% of hvac remotes use this timing
    STARTFRAME = [3500, 1750]
    ENDFRAME = [435, 10000]
    MARK = 435
    SPACE0 = 435
    SPACE1 = 1300

    def __init__(self):
        self.brand = "Generic"
        self.model = "Generic"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "fan", "dry"],
            "temperature": [x for x in range(18, 30)],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {}
        self.status = {"mode": "auto", "temperature": 25}
        self.temperature_step = 1.0

        self.to_set = {}
        # Specify wether the bits order has to be swapped
        self.is_msb = False
        # Names of 'functions' used by this object
        self.functions = []

    def get_timing(self):
        return {
            "start frame": self.STARTFRAME,
            "end frame": self.ENDFRAME,
            "mark": self.MARK,
            "space 0": self.SPACE0,
            "space 1": self.SPACE1,
        }

    def set_value(self, name, value):
        try:
            xx = getattr(self, "set_" + name)(value)
        except:
            # Does no exisat. Ignore
            pass

    def update_status(self):
        for x, y in self.to_set.items():
            self.status[x] = y
        self.to_set = {}

    def build_ircode(self):
        frames = self._build_ircode()
        if self.is_msb:
            newframes = []
            for f in frames:
                newframes.append(bytearray([bit_reverse(x) for x in f]))
            frames = newframes
        # print("Frame with msb {} are:".format(self.is_msb))
        # for f in frames:
        # print(["0x%02x"%x for x in f])
        return frames



    def to_lirc(self, frames):
        """Transform a list of frames into a LIRC compatible list of pulse timing pairs"""
        lircframe = []
        for frame in frames:
            lircframe += self.STARTFRAME
            for x in frame:
                idx = 0x80
                while idx:
                    lircframe.append(self.MARK)
                    if x & idx:
                        lircframe.append(self.SPACE1)
                    else:
                        lircframe.append(self.SPACE0)
                    idx >>= 1
            lircframe += self.ENDFRAME
        return lircframe

    def to_broadlink(self, pulses):
        array = bytearray()

        for pulse in pulses:
            pulse = round(pulse * 269 / 8192)  # 32.84ms units

            if pulse < 256:
                array += bytearray(struct.pack(">B", pulse))  # big endian (1-byte)
            else:
                array += bytearray([0x00])  # indicate next number is 2-bytes
                array += bytearray(struct.pack(">H", pulse))  # big endian (2-bytes)

        packet = bytearray([0x26, 0x00])  # 0x26 = IR, 0x00 = no repeats
        packet += bytearray(struct.pack("<H", len(array)))  # little endian byte count
        packet += array
        packet += bytearray([0x0D, 0x05])  # IR terminator

        # Add 0s to make ultimate packet size a multiple of 16 for 128-bit AES encryption.
        remainder = (
            len(packet) + 4
        ) % 16  # rm.send_data() adds 4-byte header (02 00 00 00)
        if remainder:
            packet += bytearray(16 - remainder)

        return packet


class GenPluginObject(object):
    MODELS = {"generic": HVAC}

    def __init__(self):
        self.brand = "Unknown"

    def factory(self, model):
        if model not in self.MODELS:
            return self.MODELS["generic"]()
        return self.MODELS[model]()



#!/usr/bin/python3

from setuptools import setup, find_packages

with open("pyhvac/__init__.py") as f:
    exec(f.read())

with open("README", "r") as fh:
    long_description = fh.read()

setup(
    name="pyhvac",
    version=__version__,
    author="François Wautier",
    author_email="francois@wautier.eu",
    url="http://github.com/frawau/pyhvac",
    description="Python liberary/Utility to generate A/C IR signals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    entry_points={
        "console_scripts": [
            "gaccode = pyhvac.__main__:main",
            "gcpanasonic = pyhvac.plugins.panasonic:main",
            "gcdaikin = pyhvac.plugins.daikin:main",
            "gclg = pyhvac.plugins.lg:main",
            "gcsharp = pyhvac.plugins.sharp:main",
        ],
    },
    packages=["pyhvac", "pyhvac.plugins"],
)



#! /usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Plugin to generate Panasonic AC IR commands
#
# This module is in part based on the work/code from:
#      Scott Kyle https://gist.github.com/appden/42d5272bf128125b019c45bc2ed3311f
#      mat_fr     https://www.instructables.com/id/Reverse-engineering-of-an-Air-Conditioning-control/
#      user two, mathieu, vincent
#                 https://www.analysir.com/blog/2014/12/27/reverse-engineering-panasonic-ac-infrared-protocol/
#
# Copyright (c) 2023 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
##

import struct

from .hvaclib import HVAC, GenPluginObject, bit_reverse


class Panasonic(HVAC):
    """Generic Panasonic HVAC object. It must have, at the very minimum
    "mode" and "temperature" capabilities"""

    FHEADER = b"\x40\x04\x07\x20\x00"
    F1BODY = b"\x00\x00"
    F2COMMON1 = b"\x00\x00\x70\x07"
    F2COMMON2 = b"\x00\x91\x00"
    FODOUR = b"\x40\x04\x07\x20\x01\xd9\x4c"
    FECON = b"\x40\x04\x07\x20\x01\xa1\xac"
    FILLER = b"\x01"

    def __init__(self):
        super().__init__()
        self.brand = "Panasonic"
        self.model = "Generic"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "fan", "dry"],
            "temperature": [x for x in range(16, 32)],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {}
        self.status = {"mode": "off", "temperature": 25}

        self.to_set = {}
        # Specify wether the bits order has to be swapped
        self.is_msb = False
        self.base_temp = 16
        self.functions = []

    def set_temperature(self, temp):
        if temp < self.capabilities["temperature"][0]:
            temp = self.capabilities["temperature"][0]
        elif temp > self.capabilities["temperature"][-1]:
            temp = self.capabilities["temperature"][-1]
        self.to_set["temperature"] = temp

    def code_temperature(self):
        if "temperature" in self.to_set:
            temp = self.to_set["temperature"]
            if "mode" in self.to_set and self.to_set["mode"] == "fan":
                temp = 27
        else:
            temp = self.status["temperature"]
        t = 0x20 + ((temp - self.base_temp) << 1)
        return bit_reverse(t).to_bytes(1, "big")

    def set_mode(self, mode):
        if mode not in self.capabilities["mode"]:
            mode = "auto"
        self.to_set["mode"] = mode
        if mode == "fan":
            self.to_set["temperature"] = 27
        if mode == "off":
            self.to_set = {"mode": "off"}

    def code_mode(self):
        if "mode" in self.to_set:
            mode = self.to_set["mode"]
        else:
            mode = self.status["mode"]

        if mode == "off":
            return b"\x10"
        elif mode == "dry":
            return b"\x94"
        elif mode == "fan":
            return b"\x96"
        elif mode == "cool":
            return b"\x9c"
        elif mode == "heat":
            return b"\x92"
        else:
            # If we do not know, also auto mode
            return b"\x90"

    def set_fan(self, mode):
        if "fan" not in self.capabilities:
            return
        if mode not in self.capabilities["fan"]:
            return
        self.to_set["fan"] = mode

    def code_fan(self):
        if "fan" in self.to_set:
            mode = self.to_set["fan"]
        else:
            if "fan" in self.capabilities:
                mode = self.status["fan"]
            else:
                mode = None

        if mode == "auto":
            return b"\x05"
        elif mode == "highest":
            return b"\x0e"
        elif mode == "high":
            return b"\x06"
        elif mode == "medium":
            return b"\x0a"
        elif mode == "low":
            return b"\x02"
        elif mode == "lowest":
            return b"\x0c"
        else:
            # If we do not know.... 0
            return b"\x00"

    def set_swing(self, mode):
        if "swing" not in self.capabilities:
            return
        if mode not in self.capabilities["swing"]:
            if "auto" in self.capabilities["swing"] and "auto" in mode:
                self.to_set["swing"] = "auto"
            return
        self.to_set["swing"] = mode

    def code_swing(self):
        if "swing" in self.to_set:
            mode = self.to_set["swing"]
        else:
            if "swing" in self.capabilities:
                mode = self.status["swing"]
            else:
                mode = None

        if mode == "auto":
            return b"\xf0"
        elif mode == "auto high":
            return b"\x70"
        elif mode == "auto low":
            return b"\xb0"
        elif mode == "90":
            return b"\x80"
        elif mode == "60":
            return b"\x40"
        elif mode == "45":
            return b"\xc0"
        elif mode == "30":
            return b"\x20"
        elif mode == "15":
            return b"\xa0"
        else:
            # If we do not know....0
            return b"\x00"

    def set_profile(self, mode):
        if "profile" not in self.capabilities:
            return
        if mode not in self.capabilities["profile"]:
            return
        self.to_set["profile"] = mode

    def code_profile(self):
        if "profile" in self.to_set:
            mode = self.to_set["profile"]
        else:
            if "profile" in self.capabilities:
                mode = self.status["profile"]
            else:
                mode = None
        if mode == "normal":
            return b"\x08"
        elif mode == "boost":
            return b"\x88"
        elif mode == "quiet":
            return b"\x0c"

        return b"\x00"

    def set_purifier(self, mode="off"):
        if "purifier" not in self.capabilities:
            return
        if mode not in self.capabilities["purifier"]:
            return
        self.to_set["purifier"] = mode

    def code_purifier(self):

        if "purifier" in self.to_set:
            mode = self.to_set["purifier"]
        else:
            if "purifier" in self.capabilities:
                mode = self.status["purifier"]
            else:
                mode = None

        if mode == "on":
            return b"\x20"
        else:
            return b"\x00"

    def set_cleaning(self, mode="off"):
        if "cleaning" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["cleaning"]:
            return
        # This is a toggling value AFAIK
        if self.status["cleaning"] != mode:
            self.to_set["cleaning"] = mode

    def code_cleaning(self):
        frames = []
        if "cleaning" in self.to_set:
            frames += [self.FHEADER + self.F1BODY]
            frames += [self.FODOUR]
        return frames

    def set_economy(self, mode="off"):
        if "economy" not in self.xtra_capabilities:
            return
        if mode not in self.xtra_capabilities["economy"]:
            return
        # This is a toggling value AFAIK
        if self.status["economy"] != mode:
            self.to_set["economy"] = mode

    def code_economy(self):
        frames = []
        if "economy" in self.to_set:
            frames += [self.FHEADER + self.F1BODY]
            frames += [self.FECON]
        return frames

    def crc(self, frame):
        crc = 0
        for x in frame:
            # print("Adding 0x%02x as 0x%02x"%(x,bit_reverse(x)))
            crc += bit_reverse(x)
            # print("crc now 0x%02x"%crc)
        return bit_reverse(crc & 0xFF).to_bytes(1, "big")

    def build_code(self):
        if "mode" not in self.to_set and self.status["mode"] == "fan":
            self.to_set["temperature"] = 27
        frames = [self.FHEADER + self.F1BODY]
        f2 = self.FHEADER + self.code_mode() + self.code_temperature() + self.FILLER
        f2 += (self.code_fan()[0] + self.code_swing()[0]).to_bytes(1, "big")
        f2 += self.F2COMMON1
        f2 += self.code_profile()
        f2 += self.F2COMMON2
        f2 += self.code_purifier()
        frames += [f2]
        return frames

    def _build_ircode(self):
        frames = []
        frames += self.code_cleaning()
        frames += self.code_economy()
        frames += self.build_code()
        idx = 0
        for x in frames:
            frames[idx] += self.crc(x)
            idx += 1
        return frames


class PanaCassette(Panasonic):
    """PX2T5 amd similar Panasonic HVAC object."""

    def __init__(self):
        super().__init__()
        self.model = "4 Way Cassette"
        self.capabilities = {
            "mode": ["off", "auto", "cool", "fan", "dry"],
            "temperature": [x for x in range(16, 32)],
            "fan": ["auto", "highest", "medium", "lowest"],
            "swing": ["auto", "auto high", "auto low", "90", "60", "45", "30"],
            "purifier": ["off", "on"],
        }
        # For functions that require their own frames
        self.xtra_capabilities = {"economy": ["off", "on"], "cleaning": ["off", "on"]}
        self.status = {
            "mode": "off",
            "temperature": 25,
            "fan": "auto",
            "swing": "auto",
            "purifier": "off",
            "economy": "off",
            "cleaning": "off",
        }


class PluginObject(GenPluginObject):
    MODELS = {"generic": Panasonic, "4 way cassette": PanaCassette}

    def __init__(self):
        self.brand = "panasonic"


def main():
    import argparse
    import base64

    parser = argparse.ArgumentParser(description="Generate Panasonic A/C codes.")
    # version="%prog " + __version__ + "/" + bl.__version__)
    parser.add_argument(
        "-M",
        "--model",
        type=str,
        default="generic",
        help="Set the A/C model. (default generic).",
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="List known models and return.",
    )
    parser.add_argument(
        "-t", "--temp", type=int, default=25, help="Temperature (°C). (default 25)."
    )
    parser.add_argument(
        "-m",
        "--mode",
        choices=["auto", "cool", "dry", "fan", "off"],
        default="auto",
        help="Mode, one of 'off', 'auto', 'cool', 'dry' or 'fan'. (default 'auto').",
    )
    parser.add_argument(
        "-f",
        "--fan",
        choices=["auto", "high", "middle", "low"],
        default="auto",
        help="Fan, one of 'auto', 'high', 'middle' or 'low'. (default 'auto').",
    )
    parser.add_argument(
        "-s",
        "--swing",
        choices=["auto", "auto high", "auto low", "90", "60", "45", "30"],
        default="auto",
        help="Swing, one of 'auto','auto high','auto low', '90', '60', '45', '30'. (default 'auto').",
    )
    parser.add_argument(
        "-n", "--nanoex", action="store_true", default=False, help="nanoeX mode"
    )
    parser.add_argument(
        "-o", "--odour", action="store_true", default=False, help="Odour Wash mode"
    )
    parser.add_argument(
        "-e", "--economy", action="store_true", default=False, help="Economy mode"
    )
    parser.add_argument(
        "-l",
        "--lirc",
        action="store_true",
        default=False,
        help="Output LIRC compatible timing",
    )
    parser.add_argument(
        "-b",
        "--broadlink",
        action="store_true",
        default=False,
        help="Output Broadlink timing",
    )
    parser.add_argument(
        "-B",
        "--base64",
        action="store_true",
        default=False,
        help="Output Broadlink timing base64 encoded",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))

    if opts.list:
        print(f"Available models are: {[x for x in PluginObject().MODELS.keys()]}")

    device = PluginObject().factory(opts.model)
    frames = []
    device.set_temperature(opts.temp)
    device.set_fan(opts.fan)
    device.set_swing(opts.swing)
    device.set_purifier((opts.nanoex and "on") or "off")
    device.set_cleaning((opts.odour and "on") or "off")
    device.set_economy((opts.economy and "on") or "off")
    device.set_mode(opts.mode)

    frames = device.build_ircode()

    if opts.lirc:
        lircf = device.to_lirc(frames)
        while lircf:
            print("\t".join(["%d" % x for x in lircf[:6]]))
            lircf = lircf[6:]
    elif opts.broadlink or opts.base64:
        lircf = device.to_lirc(frames)
        bframe = device.to_broadlink([int(x) for x in lircf])
        if opts.base64:
            print("{}".format(str(base64.b64encode(bframe), "ascii")))
        else:
            print("{}".format(bframe.hex()))
    else:
        for f in frames:
            print(" ".join(["%02x" % x for x in f]))


if __name__ == "__main__":
    main()
